@@ -9,10 +9,18 @@ extern "C" {
 #endif
 
 #define PD_MAJOR_VERSION 0
-#define PD_MINOR_VERSION 50
+#define PD_MINOR_VERSION 55
 #define PD_BUGFIX_VERSION 1
 #define PD_TEST_VERSION ""
-extern int pd_compatibilitylevel;   /* e.g., 43 for pd 0.43 compatibility */
+
+/* compile-time version check:
+   #if PD_VERSION_CODE < PD_VERSION(0, 56, 0)
+      // put legacy code for Pd<<0.56 in here
+   #endif
+ */
+#define PD_VERSION(major, minor, bugfix) \
+    (((major) << 16) + ((minor) << 8) + ((bugfix) > 255 ? 255 : (bugfix)))
+#define PD_VERSION_CODE PD_VERSION(PD_MAJOR_VERSION, PD_MINOR_VERSION, PD_BUGFIX_VERSION)
 
 /* old name for "MSW" flag -- we have to take it for the sake of many old
 "nmakefiles" for externs, which will define NT and not MSW */
@@ -29,6 +37,7 @@ extern int pd_compatibilitylevel;   /* e.g., 43 for pd 0.43 compatibility */
 #endif /* _MSC_VER */
 
     /* the external storage class is "extern" in UNIX; in MSW it's ugly. */
+#ifndef EXTERN
 #ifdef _WIN32
 #ifdef PD_INTERNAL
 #define EXTERN __declspec(dllexport) extern
@@ -38,18 +47,32 @@ extern int pd_compatibilitylevel;   /* e.g., 43 for pd 0.43 compatibility */
 #else
 #define EXTERN extern
 #endif /* _WIN32 */
+#endif /* EXTERN */
 
     /* On most c compilers, you can just say "struct foo;" to declare a
-    structure whose elements are defined elsewhere.  On MSVC, when compiling
-    C (but not C++) code, you have to say "extern struct foo;".  So we make
-    a stupid macro: */
+    structure whose elements are defined elsewhere.  On very old MSVC versions,
+    when compiling C (but not C++) code, you have to say "extern struct foo;".
+    So we make a stupid macro: */
 #if defined(_MSC_VER) && !defined(_LANGUAGE_C_PLUS_PLUS) \
-    && !defined(__cplusplus)
+    && !defined(__cplusplus) && (_MSC_VER < 1700)
 #define EXTERN_STRUCT extern struct
 #else
 #define EXTERN_STRUCT struct
 #endif
 
+
+/* util for better inlining of static functions in headers */
+#if defined(__cplusplus)
+# define PD_INLINE inline
+#else
+# if (__STDC_VERSION__ >= 199901L)
+#  define PD_INLINE static inline
+# else
+#  define PD_INLINE static
+# endif
+#endif
+
+
 /* Define some attributes, specific to the compiler */
 #if defined(__GNUC__)
 #define ATTRIBUTE_FORMAT_PRINTF(a, b) __attribute__ ((format (printf, a, b)))
@@ -57,12 +80,37 @@ extern int pd_compatibilitylevel;   /* e.g., 43 for pd 0.43 compatibility */
 #define ATTRIBUTE_FORMAT_PRINTF(a, b)
 #endif
 
+#if __STDC_VERSION__ >= 201112L
+#include <assert.h>
+#define PD_STATIC_ASSERT _Static_assert
+#elif __cplusplus >= 201103L
+#define PD_STATIC_ASSERT static_assert
+#else
+#define PD_STATIC_ASSERT(condition, message) /* no-op */
+#endif
+
+/* deprecation warning */
+#ifndef PD_DEPRECATED
+# ifdef __GNUC__
+#  define PD_DEPRECATED __attribute__ ((deprecated))
+# elif defined(_MSC_VER) && _MSC_VER >= 1300
+#  define PD_DEPRECATED __declspec(deprecated)
+# else
+#  define PD_DEPRECATED
+#  pragma message("PD_DEPRECATED not defined for this compiler")
+# endif
+#endif
+
 #if !defined(_SIZE_T) && !defined(_SIZE_T_)
 #include <stddef.h>     /* just for size_t -- how lame! */
 #endif
 
-/* Microsoft Visual Studio is not C99, it does not provide stdint.h */
-#ifdef _MSC_VER
+#include <stddef.h> /* for size_t and offsetof */
+
+/* Microsoft Visual Studio is not C99, but since VS2015 has included most C99 headers:
+   https://docs.microsoft.com/en-us/previous-versions/hh409293(v=vs.140)#c-runtime-library
+   These definitions recreate stdint.h types, but only in pre-2015 Visual Studio: */
+#if defined(_MSC_VER) && _MSC_VER < 1900
 typedef signed __int8     int8_t;
 typedef signed __int16    int16_t;
 typedef signed __int32    int32_t;
@@ -78,12 +126,14 @@ typedef unsigned __int64  uint64_t;
 /* for FILE, needed by sys_fopen() and sys_fclose() only */
 #include <stdio.h>
 
+EXTERN int pd_compatibilitylevel;   /* e.g., 43 for pd 0.43 compatibility */
+
 #define MAXPDSTRING 1000        /* use this for anything you want */
 #define MAXPDARG 5              /* max number of args we can typecheck today */
 
 /* signed and unsigned integer types the size of a pointer:  */
 #if !defined(PD_LONGINTTYPE)
-#if defined(_WIN32) && defined(__x86_64__)
+#if defined(_WIN32) && defined(_WIN64)
 #define PD_LONGINTTYPE long long
 #else
 #define PD_LONGINTTYPE long
@@ -98,11 +148,11 @@ typedef unsigned __int64  uint64_t;
 #if PD_FLOATSIZE == 32
 # define PD_FLOATTYPE float
 /* an unsigned int of the same size as FLOATTYPE: */
-# define PD_FLOATUINTTYPE unsigned int
+# define PD_FLOATUINTTYPE uint32_t
 
 #elif PD_FLOATSIZE == 64
 # define PD_FLOATTYPE double
-# define PD_FLOATUINTTYPE unsigned long
+# define PD_FLOATUINTTYPE uint64_t
 #else
 # error invalid FLOATSIZE: must be 32 or 64
 #endif
@@ -164,7 +214,7 @@ typedef union word
 
 typedef enum
 {
-    A_NULL,
+    A_NULL = 0,
     A_FLOAT,
     A_SYMBOL,
     A_POINTER,
@@ -279,7 +329,7 @@ EXTERN t_symbol *gensym(const char *s);
 EXTERN t_gotfn getfn(const t_pd *x, t_symbol *s);
 EXTERN t_gotfn zgetfn(const t_pd *x, t_symbol *s);
 EXTERN void nullfn(void);
-EXTERN void pd_vmess(t_pd *x, t_symbol *s, char *fmt, ...);
+EXTERN void pd_vmess(t_pd *x, t_symbol *s, const char *fmt, ...);
 
 /* the following macros are for sending non-type-checkable messages, i.e.,
 using function lookup but circumventing type checking on arguments.  Only
@@ -374,7 +424,7 @@ EXTERN void clock_delay(t_clock *x, double delaytime);
 EXTERN void clock_unset(t_clock *x);
 EXTERN void clock_setunit(t_clock *x, double timeunit, int sampflag);
 EXTERN double clock_getlogicaltime(void);
-EXTERN double clock_getsystime(void); /* OBSOLETE; use clock_getlogicaltime() */
+PD_DEPRECATED EXTERN double clock_getsystime(void); /* use clock_getlogicaltime() */
 EXTERN double clock_gettimesince(double prevsystime);
 EXTERN double clock_gettimesincewithunits(double prevsystime,
     double units, int sampflag);
@@ -445,6 +495,8 @@ EXTERN int sys_fontheight(int fontsize);
 EXTERN void canvas_dataproperties(t_glist *x, t_scalar *sc, t_binbuf *b);
 EXTERN int canvas_open(const t_canvas *x, const char *name, const char *ext,
     char *dirresult, char **nameresult, unsigned int size, int bin);
+EXTERN t_float canvas_getsr(t_canvas *x);
+EXTERN int canvas_getsignallength(t_canvas *x);
 
 /* ---------------- widget behaviors ---------------------- */
 
@@ -458,13 +510,35 @@ EXTERN const t_parentwidgetbehavior *pd_getparentwidget(t_pd *x);
 /* -------------------- classes -------------- */
 
 #define CLASS_DEFAULT 0         /* flags for new classes below */
-#define CLASS_PD 1
-#define CLASS_GOBJ 2
-#define CLASS_PATCHABLE 3
-#define CLASS_NOINLET 8
-
+#define CLASS_PD 1              /* non-canvasable (bare) pd such as an inlet */
+#define CLASS_GOBJ 2            /* pd that can belong to a canvas */
+#define CLASS_PATCHABLE 3       /* pd that also can have inlets and outlets */
 #define CLASS_TYPEMASK 3
 
+#define CLASS_NOINLET 8             /* suppress left inlet */
+#define CLASS_MULTICHANNEL 0x10     /* can deal with multichannel sigs */
+#define CLASS_NOPROMOTESIG 0x20     /* don't promote scalars to signals */
+#define CLASS_NOPROMOTELEFT 0x40    /* not even the main (left) inlet */
+
+/*
+    Setting a tilde object's CLASS_MULTICHANNEL flag declares that it can
+    deal with multichannel inputs.  In this case the channel counts of
+    the inputs might not match; it's up to the dsp method to figure out what
+    to do.  Also, the output signal vectors aren't allocated.  The output
+    channel counts have to be specified by the object at DSP time.  If
+    the object can't put itself on the DSP chain it then has to create
+    outputs anyway and arrange to zero them.
+
+    By default, if a tilde object's inputs are unconnected, Pd fills them
+    in by adding scalar-to-vector conversions to the DSP chain as needed before
+    calling the dsp method.  This behavior can be suppressed for the left
+    (main) inlet by setting CLASS_NOPROMOTELEFT and for one or more non-main
+    inlets by setting CLASS_NOPROMOTESIG.  Seeing this, the object can then
+    opt to supply a faster routine; for example, "+" can do a vector-scalar
+    add.  In any case, signal outputs are all vectors, and are allocated
+    automatically unless the CLASS_MULTICHANNEL flag is also set.
+*/
+
 EXTERN t_class *class_new(t_symbol *name, t_newmethod newmethod,
     t_method freemethod, size_t size, int flags, t_atomtype arg1, ...);
 
@@ -473,7 +547,7 @@ EXTERN t_class *class_new64(t_symbol *name, t_newmethod newmethod,
 
 EXTERN void class_free(t_class *c);
 
-#if PDINSTANCE
+#ifdef PDINSTANCE
 EXTERN t_class *class_getfirst(void);
 #endif
 
@@ -495,10 +569,11 @@ EXTERN const char *class_gethelpname(const t_class *c);
 EXTERN const char *class_gethelpdir(const t_class *c);
 EXTERN void class_setdrawcommand(t_class *c);
 EXTERN int class_isdrawcommand(const t_class *c);
-EXTERN void class_domainsignalin(t_class *c, int onset);
 EXTERN void class_set_extern_dir(t_symbol *s);
+EXTERN void class_domainsignalin(t_class *c, int onset);
 #define CLASS_MAINSIGNALIN(c, type, field) \
-    class_domainsignalin(c, (char *)(&((type *)0)->field) - (char *)0)
+    PD_STATIC_ASSERT(sizeof(((type *)NULL)->field) == sizeof(t_float), "field must be t_float!"); \
+    class_domainsignalin(c, offsetof(type, field))
 
          /* prototype for functions to save Pd's to a binbuf */
 typedef void (*t_savefn)(t_gobj *x, t_binbuf *b);
@@ -528,21 +603,30 @@ EXTERN void class_setfreefn(t_class *c, t_classfreefn fn);
 #endif
 
 /* ------------   printing --------------------------------- */
+
 EXTERN void post(const char *fmt, ...);
 EXTERN void startpost(const char *fmt, ...);
 EXTERN void poststring(const char *s);
 EXTERN void postfloat(t_floatarg f);
 EXTERN void postatom(int argc, const t_atom *argv);
 EXTERN void endpost(void);
-EXTERN void error(const char *fmt, ...) ATTRIBUTE_FORMAT_PRINTF(1, 2);
-EXTERN void verbose(int level, const char *fmt, ...) ATTRIBUTE_FORMAT_PRINTF(2, 3);
+
 EXTERN void bug(const char *fmt, ...) ATTRIBUTE_FORMAT_PRINTF(1, 2);
 EXTERN void pd_error(const void *object, const char *fmt, ...) ATTRIBUTE_FORMAT_PRINTF(2, 3);
-EXTERN void logpost(const void *object, const int level, const char *fmt, ...)
+
+/* for logpost(); does *not* work with verbose()! */
+typedef enum {
+    PD_CRITICAL = 0,
+    PD_ERROR,
+    PD_NORMAL,
+    PD_DEBUG,
+    PD_VERBOSE
+} t_loglevel;
+
+EXTERN void logpost(const void *object, int level, const char *fmt, ...)
     ATTRIBUTE_FORMAT_PRINTF(3, 4);
-EXTERN void sys_logerror(const char *object, const char *s);
-EXTERN void sys_unixerror(const char *object);
-EXTERN void sys_ouch(void);
+
+PD_DEPRECATED EXTERN void verbose(int level, const char *fmt, ...) ATTRIBUTE_FORMAT_PRINTF(2, 3); /* avoid this: use logpost() instead */
 
 
 /* ------------  system interface routines ------------------- */
@@ -582,22 +666,38 @@ typedef union _sampleint_union {
 
 typedef struct _signal
 {
-    int s_n;            /* number of points in the array */
-    t_sample *s_vec;    /* the array */
-    t_float s_sr;         /* sample rate */
-    int s_refcount;     /* number of times used */
-    int s_isborrowed;   /* whether we're going to borrow our array */
+    union
+    {
+        int s_length;       /* number of items per channel */
+        int s_n;            /* for source compatibility: pre-0.54 name */
+    };
+    t_sample *s_vec;        /* the samples, s_nchans vectors of s_length */
+    t_float s_sr;           /* samples per second per channel */
+    int s_nchans;           /* number of channels */
+    int s_overlap;          /* number of times each sample appears */
+    int s_refcount;         /* number of times signal is referenced */
+    int s_isborrowed;       /* whether we're going to borrow our array */
+    int s_isscalar;         /* scalar for an unconnected signal input */
     struct _signal *s_borrowedfrom;     /* signal to borrow it from */
     struct _signal *s_nextfree;         /* next in freelist */
     struct _signal *s_nextused;         /* next in used list */
-    int s_vecsize;      /* allocated size of array in points */
+    int s_nalloc;      /* allocated size of array in points */
 } t_signal;
 
 typedef t_int *(*t_perfroutine)(t_int *args);
 
+EXTERN t_signal *signal_new(int length, int nchans, t_float sr,
+    t_sample *scalarptr);
+EXTERN void signal_setmultiout(t_signal **sig, int nchans);
+
 EXTERN t_int *plus_perform(t_int *args);
+EXTERN t_int *plus_perf8(t_int *args);
 EXTERN t_int *zero_perform(t_int *args);
+EXTERN t_int *zero_perf8(t_int *args);
 EXTERN t_int *copy_perform(t_int *args);
+EXTERN t_int *copy_perf8(t_int *args);
+EXTERN t_int *scalarcopy_perform(t_int *args);
+EXTERN t_int *scalarcopy_perf8(t_int *args);
 
 EXTERN void dsp_add_plus(t_sample *in1, t_sample *in2, t_sample *out, int n);
 EXTERN void dsp_add_copy(t_sample *in, t_sample *out, int n);
@@ -620,10 +720,6 @@ EXTERN void mayer_ifft(int n, t_sample *real, t_sample *imag);
 EXTERN void mayer_realfft(int n, t_sample *real);
 EXTERN void mayer_realifft(int n, t_sample *real);
 
-EXTERN float *cos_table;
-#define LOGCOSTABSIZE 9
-#define COSTABSIZE (1<<LOGCOSTABSIZE)
-
 EXTERN int canvas_suspend_dsp(void);
 EXTERN void canvas_resume_dsp(int oldstate);
 EXTERN void canvas_update_dsp(void);
@@ -632,7 +728,7 @@ EXTERN int canvas_dspstate;
 /*   up/downsampling */
 typedef struct _resample
 {
-  int method;       /* up/downsampling method ID */
+  int method;       /* unused */
 
   int downsample; /* downsampling factor */
   int upsample;   /* upsampling factor */
@@ -676,12 +772,12 @@ EXTERN_STRUCT _garray;
 #define t_garray struct _garray
 
 EXTERN t_class *garray_class;
-EXTERN int garray_getfloatarray(t_garray *x, int *size, t_float **vec);
+PD_DEPRECATED EXTERN int garray_getfloatarray(t_garray *x, int *size, t_float **vec); /* use garray_getfloatwords() */
 EXTERN int garray_getfloatwords(t_garray *x, int *size, t_word **vec);
 EXTERN void garray_redraw(t_garray *x);
 EXTERN int garray_npoints(t_garray *x);
 EXTERN char *garray_vec(t_garray *x);
-EXTERN void garray_resize(t_garray *x, t_floatarg f);  /* avoid; use this: */
+PD_DEPRECATED EXTERN void garray_resize(t_garray *x, t_floatarg f); /* use garray_resize_long() */
 EXTERN void garray_resize_long(t_garray *x, long n);   /* better version */
 EXTERN void garray_usedindsp(t_garray *x);
 EXTERN void garray_setsaveit(t_garray *x, int saveit);
@@ -697,14 +793,62 @@ EXTERN int value_setfloat(t_symbol *s, t_float f);
 /* ------- GUI interface - functions to send strings to TK --------- */
 typedef void (*t_guicallbackfn)(t_gobj *client, t_glist *glist);
 
-EXTERN void sys_vgui(char *fmt, ...);
-EXTERN void sys_gui(char *s);
+PD_DEPRECATED EXTERN void sys_vgui(const char *fmt, ...); /* avoid this: use pdgui_vmess() instead */
+PD_DEPRECATED EXTERN void sys_gui(const char *s); /* avoid this: use pdgui_vmess() instead */
+
 EXTERN void sys_pretendguibytes(int n);
 EXTERN void sys_queuegui(void *client, t_glist *glist, t_guicallbackfn f);
 EXTERN void sys_unqueuegui(void *client);
     /* dialog window creation and destruction */
-EXTERN void gfxstub_new(t_pd *owner, void *key, const char *cmd);
-EXTERN void gfxstub_deleteforkey(void *key);
+PD_DEPRECATED EXTERN void gfxstub_new(t_pd *owner, void *key, const char *cmd); /* avoid this: use pdgui_stub_vnew() instead */
+PD_DEPRECATED EXTERN void gfxstub_deleteforkey(void *key); /* avoid this: use pdgui_stub_deleteforkey() instead */
+
+/*
+ * send a message to the GUI, with a simplified formatting syntax
+ * <destination>: receiver on the GUI side (e.g. a Tcl/Tk 'proc')
+ * <fmt>: string of format specifiers
+ * <...>: values according to the format specifiers
+ *
+ * the <destination> can be a NULL pointer (in which case it is ignored)
+ * the user of NULL as a <destination> is discouraged
+
+ * depending on the format specifiers, one or more values are passed
+ *    'f' : <double:value>        : a floating point number
+ *    'i' : <int:value>           : an integer number
+ *    's' : <const char*:value>   : a string
+ *    'r' : <const char*:value>   : a raw string
+ *    'x' : <void*:value>         : a generic pointer
+ *    'o' : <t_object*:value>     : an graphical object
+ *    '^' : <t_canvas*:value>     : a toplevel window (legacy)
+ *    'c' : <t_canvas*:value>     : a canvas (on a window)
+ *    'F' : <int:size> <const t_float*:values>: array of t_float's
+ *    'S' : <int:size> <const char**:values>: array of strings
+ *    'R' : <int:size> <const char**:values>: array of raw strings
+ *    'a' : <int:size> <const t_atom*:values>: list of atoms
+ *    'A' : <int:size> <const t_atom*:values>: array of atoms
+ *    'w' : <int:size> <const t_word*:values>: list of floatwords
+ *    'W' : <int:size> <const t_word*:values>: array of floatwords
+ *    'm' : <t_symbol*s:recv> <int:argc> <t_atom*:argv>: a Pd message
+ *    'p' : <int:size> <const char*:values>  : a pascal string (explicit size; not \0-terminated)
+ *    'k' : <int:color>           : a color (or kolor, if you prefer)
+ *    ' ' : none                  : ignored
+ * the use of the specifiers 'x^' is discouraged
+ * raw-strings ('rR') should only be used for constant, well-known strings
+ */
+EXTERN void pdgui_vmess(const char* destination, const char* fmt, ...);
+
+
+/* improved dialog window creation
+ * this will insert a first argument to <destination> based on <key>
+ * which the GUI can then use to callback.
+ * gfxstub_new() ensures that the given receiver will be available,
+ * even if the <owner> has been removed in the meantime.
+ * see pdgui_vmess() for a description of <fmt> and the varargs
+ */
+
+EXTERN void pdgui_stub_vnew(t_pd *owner, const char* destination, void *key, const char* fmt, ...);
+EXTERN void pdgui_stub_deleteforkey(void *key);
+
 
 extern t_class *glob_pdobject;  /* object to send "pd" messages */
 
@@ -730,26 +874,29 @@ defined, there is a "te_xpix" field in objects, not a "te_xpos" as before: */
 
 #define PD_USE_TE_XPIX
 
-#ifndef _MSC_VER /* Microoft compiler can't handle "inline" function/macros */
-#if defined(__i386__) || defined(__x86_64__) || defined(__arm__)
-/* a test for NANs and denormals.  Should only be necessary on i386. */
+#if defined(__i386__) || defined(__x86_64__) || defined(__arm__) || defined(__aarch64__)
+
+/* a test for NANs and denormals.
+   WARNING: PD_BADFLOAT(0) = 1 -- you can replace a 'BADFLOAT' by 0 but don't
+just do nothing with it. */
+
 #if PD_FLOATSIZE == 32
 
-typedef  union
+typedef union
 {
     t_float f;
     unsigned int ui;
-}t_bigorsmall32;
+} t_bigorsmall32;
 
-static inline int PD_BADFLOAT(t_float f)  /* malformed float */
+PD_INLINE int PD_BADFLOAT(t_float f)  /* malformed float */
 {
     t_bigorsmall32 pun;
     pun.f = f;
     pun.ui &= 0x7f800000;
-    return((pun.ui == 0) | (pun.ui == 0x7f800000));
+    return((f != 0) && ((pun.ui == 0) | (pun.ui == 0x7f800000)));
 }
 
-static inline int PD_BIGORSMALL(t_float f)  /* exponent outside (-64,64) */
+PD_INLINE int PD_BIGORSMALL(t_float f)  /* exponent outside (-64,64) */
 {
     t_bigorsmall32 pun;
     pun.f = f;
@@ -758,21 +905,21 @@ static inline int PD_BIGORSMALL(t_float f)  /* exponent outside (-64,64) */
 
 #elif PD_FLOATSIZE == 64
 
-typedef  union
+typedef union
 {
     t_float f;
     unsigned int ui[2];
-}t_bigorsmall64;
+} t_bigorsmall64;
 
-static inline int PD_BADFLOAT(t_float f)  /* malformed double */
+PD_INLINE int PD_BADFLOAT(t_float f)  /* malformed double */
 {
     t_bigorsmall64 pun;
     pun.f = f;
     pun.ui[1] &= 0x7ff00000;
-    return((pun.ui[1] == 0) | (pun.ui[1] == 0x7ff00000));
+    return((f != 0) && ((pun.ui[1] == 0) | (pun.ui[1] == 0x7ff00000)));
 }
 
-static inline int PD_BIGORSMALL(t_float f)  /* exponent outside (-512,512) */
+PD_INLINE int PD_BIGORSMALL(t_float f)  /* exponent outside (-512,512) */
 {
     t_bigorsmall64 pun;
     pun.f = f;
@@ -785,21 +932,27 @@ static inline int PD_BIGORSMALL(t_float f)  /* exponent outside (-512,512) */
 #define PD_BIGORSMALL(f) 0
 #endif
 
-#else   /* _MSC_VER */
-#if PD_FLOATSIZE == 32
-#define PD_BADFLOAT(f) ((((*(unsigned int*)&(f))&0x7f800000)==0) || \
-    (((*(unsigned int*)&(f))&0x7f800000)==0x7f800000))
-/* more stringent test: anything not between 1e-19 and 1e19 in absolute val */
-#define PD_BIGORSMALL(f) ((((*(unsigned int*)&(f))&0x60000000)==0) || \
-    (((*(unsigned int*)&(f))&0x60000000)==0x60000000))
-#else   /* 64 bits... don't know what to do here */
-#define PD_BADFLOAT(f) (!(((f) >= 0) || ((f) <= 0)))
-#define PD_BIGORSMALL(f) ((f) > 1e150 || (f) <  -1e150 \
-    || (f) > -1e-150 && (f) < 1e-150 )
+    /* get major/minor/bugfix version numbers and version code at run time */
+EXTERN unsigned int sys_getversion(int *major, int *minor, int *bugfix);
+#ifndef PD_INTERNAL
+/* ABI compat hack: do not use _sys_getversioncode() directly! */
+PD_INLINE unsigned int _sys_getversioncode(int*major, int*minor, int*bugfix)
+{
+    int a, b, c;
+    sys_getversion(major, minor, bugfix);
+    sys_getversion(&a, &b, &c);
+    return PD_VERSION(a, b, c);
+}
+#define sys_getversion _sys_getversioncode
 #endif
-#endif /* _MSC_VER */
-    /* get version number at run time */
-EXTERN void sys_getversion(int *major, int *minor, int *bugfix);
+
+    /* get a Pd API function pointer by name. Returns NULL if the function
+    does not exist. For example, This allows to use recently introduced API
+    functions while providing a fallback for older Pd versions. */
+EXTERN t_method sys_getfunbyname(const char *name);
+
+    /* get floatsize at run time */
+EXTERN unsigned int sys_getfloatsize(void);
 
 EXTERN_STRUCT _instancemidi;
 #define t_instancemidi struct _instancemidi
@@ -859,6 +1012,7 @@ EXTERN t_pdinstance pd_maininstance;
 #ifdef PDINSTANCE
 EXTERN t_pdinstance *pdinstance_new(void);
 EXTERN void pd_setinstance(t_pdinstance *x);
+EXTERN t_pdinstance *pd_getinstance(void);
 EXTERN void pdinstance_free(t_pdinstance *x);
 #endif /* PDINSTANCE */
 
@@ -873,7 +1027,18 @@ EXTERN void pdinstance_free(t_pdinstance *x);
 #endif
 
 #ifdef PDINSTANCE
-extern PERTHREAD t_pdinstance *pd_this;
+#ifdef _WIN32
+/* Windows does not allow exporting thread-local variables from DLLs,
+so externals need to get 'pd_this' with an (implicit) function call.
+Internally, we may directly access 'pd_this', but we must not export it! */
+#ifdef PD_INTERNAL
+extern PERTHREAD t_pdinstance *pd_this; /* not EXTERN! */
+#else
+#define pd_this pd_getinstance()
+#endif /* PD_INTERNAL */
+#else
+EXTERN PERTHREAD t_pdinstance *pd_this;
+#endif /* _WIN32 */
 EXTERN t_pdinstance **pd_instances;
 EXTERN int pd_ninstances;
 #else
@@ -901,6 +1066,21 @@ EXTERN t_symbol s_pointer, s_float, s_symbol, s_bang, s_list, s_anything,
 EXTERN t_canvas *pd_getcanvaslist(void);
 EXTERN int pd_getdspstate(void);
 
+/* x_text.c */
+EXTERN t_binbuf *text_getbufbyname(t_symbol *s); /* get binbuf from text obj */
+EXTERN void text_notifybyname(t_symbol *s);      /* notify it was modified */
+
+/* g_undo.c */
+/* store two message-sets to be sent to the object's <s> method for 'undo'ing
+ * resp. 'redo'ing the current state of an object.
+ * this creates an internal copy of the atom-lists (so the caller is responsible
+ * for freeing any dynamically allocated data)
+ * this is a no-op if called during 'undo' (resp. 'redo').
+ */
+EXTERN void pd_undo_set_objectstate(t_canvas*canvas, t_pd*x, t_symbol*s,
+                                    int undo_argc, t_atom*undo_argv,
+                                    int redo_argc, t_atom*redo_argv);
+
 #if defined(_LANGUAGE_C_PLUS_PLUS) || defined(__cplusplus)
 }
 #endif
